/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#include "minrtsFrameData.h"
#include "minrtsFftUtil.h"

#define DEBUG 1
/* #define RECAL 1 */
/**********************************************************
 * Usage:
 *    oirSci2HiPhot -AB ABdata -ref coordFile -cross crosscoeff \
 *       -outbase outFileBase 
 *
 *    data is a chopped SCI_PHOT file
 *
 *    coordFile is a fits file containing at least a detector def file
 *      defining the wavelength/distortion coefficients
 *
 *    crosscoeff is a crosscoefficient (kappa coefficient) file name
 *
 *    outFileBase is the base of the output file.  Two output
 *    files are created: outFileBaseA.fits and outFileBaseB.fits
 *    corresponding to the output of oirChopPhotometry for
 *    AOPEN and BOPEN shutters respectively.
 *
 *    
 ************************************************************* */

void Usage() {
   printf("Usage: oirSci2HiPhot -AB inFile \n");
   printf("    -outbase outFileBase\n");
   printf("    -ref refCoordFile\n");
   printf("    -cross crossCoeffFile\n");
   printf("inFile: a input file of chopped SCI_PHOT data\n");
   printf("outFileBase: beginning of output file name, the two output files\n");
   printf("are outFileBaseA.fits and outFileBaseB.fits \n");
   printf("refCoordFile: contains coordinate distortion info:\n\n");
}
/*
   float pixel copying utility
*/
void pixCpy (float *dest, float* source, long nPix) {
   memcpy (dest, source, nPix*sizeof(float));
   return;
   }  /* end pixCpy */

int main (int argc, char** argv) {

   char opt1[12];
   char gris[12];
   char shutter[12];
   float dit;

   int i,ierr;
   int nx, ny;
   long npix;
   int  ix, iy;
/*
   int NBIN; 
   int ibin;
*/
   long  nRows, iRow;
   int iFile;

   int iChan ;

   char outFileBase[130];
   char outFileName[130];

   fitsfile *inFile;

   imaging_fdata_table_def *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_detector_table_def *CoordDef;
   imaging_fdata_row *outData;
   int nRegion, iRegion;

   minrtsFrameData **inFrames;
   minrtsFrameData **outFrames[2];
   minrtsFrameData *scratch;
   minrtsFrameData *cross ;

   int crossSpecified;
   int iarg;


   float t1, t2;
   float p02, p13;

   char *bang="!";
   char *ABshutter   = "ABOPEN";
   char *sciPhot     = "SCI_PHOT";
   fitsfile *outFile = NULL;
   int one = 1;

   char *inFileName    = NULL;
   char *refFileName   = NULL;
   char *crossFileName = NULL;

   char *AFileTail   = ".Aphotometry.fits";
   char *BFileTail   = ".Bphotometry.fits";

   ierr = 0;
   outFileBase[0]  = 0;
/*
      parse input
*/
   printf("\n      ****** oirSci2HiPhot ******\n");
   if (argc <2) {
      Usage();
      return 1;
   }
   iarg   = 1;
   while (iarg < argc) {
      if (strcmp(argv[iarg], "-AB")==0) 
        inFileName = argv[++iarg];
      if (strcmp(argv[iarg], "-cross")==0) 
        crossFileName = argv[++iarg];
      if (strcmp(argv[iarg], "-outbase")==0) {
         strcpy(outFileBase, bang);
         strcat(outFileBase, argv[++iarg]);
      }
      if (strcmp(argv[iarg], "-ref")==0) 
        refFileName = argv[++iarg];
      ++iarg;
   }
   if (inFileName == NULL) {
      printf("ERROR:you need to specify an input file \n");
      return 1;
   } else {
      printf("   input File:%s\n ", inFileName);
   }
   if (refFileName == NULL) {
      printf("you have not specified a REFERENCE file with coordinate information.\n");
      printf("I will use the input data file for this information.\n");
      return 1;
   }
   if (crossFileName == NULL) {
      crossSpecified = 0;
/*      printf("ERROR:you need to specify the cross coefficient file name\n"); */
      printf("crossCoefficient file NOT SPECIFIED; using UNITY coefficients\n");
/*      return 1; */
   } else crossSpecified = 1;
   if (outFileBase[0] == 0) {
      printf("ERROR:you need to specify the output file base name\n");
      return 1;
   }
/*
       open input file and get some information
*/
   i = fits_open_file(&inFile, inFileName, READONLY, &ierr) ;
   if (ierr != 0) {
      printf("ERROR:cfitsio %i opening input file %s\n", ierr, inFileName);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
   i = fits_read_key(inFile, TINT, "NBIN", &NBIN, 0, &ierr);
      if(ierr == 0) printf("   NBIN = %i\n", NBIN);
      else { 
         printf(" NBIN is not defined in input file %s \n",inFileName);
         return 1;
      }
*/
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS SHUT NAME", 
      shutter, 0, &ierr);
   if (strncmp(shutter, ABshutter, 6)!=0) {
      printf("   I expected shutter ABOPEN not %s \n", shutter);
      printf("   I hope you are trying to compute cross-coefficients\n"); 
/*      return 1; */
   }
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", 
      opt1,0, &ierr);
   if (strcmp(opt1, sciPhot) != 0) {
      printf("ERROR:I expect opt1 to be SCI_PHOT not %s \n", opt1);
      return 1;
   }
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", 
      gris,0,&ierr);
/*
      move to detector section
*/
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr !=0) {
      printf("ERROR:%i opening detector table in %s\n", 
         ierr, inFileName);
      return 1;
   }
   nRegion = DetectorDef->nregion;
   i= fits_movrel_hdu(inFile, 1, NULL, &ierr);
   i = fits_get_num_rows (inFile, &nRows, &ierr);
   i = fits_close_file(inFile, &ierr);
   printf("   disperser is = %s\n", gris);

/*
       get detector table from file with coordinate distortion info
*/

   i = fits_open_file(&inFile, refFileName, READONLY, &ierr) ;
   if (ierr != 0) {
      printf("ERROR:cfitsio %i opening coord file %s\n", ierr, refFileName);
      printf("%s \n",strerror(errno));
      return 1;
   }
   CoordDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr !=0) {
      printf("ERROR: %i opening Coord detector table in %s\n", 
         ierr, refFileName);
      return 1;
   }
   printf("   using optical distortion data from %s\n", refFileName);
   i = fits_close_file(inFile, &ierr);

/*
          create FrameData structures to hold data
*/
   inFrames = (minrtsFrameData **)malloc (nRows * sizeof(minrtsFrameData *));
   for (iFile=0; iFile<2; iFile++) outFrames[iFile]  = 
      (minrtsFrameData **)malloc (nRows * sizeof(minrtsFrameData *));
/*
          read in output of oirChopPhotoImages 
*/
   for (iRow=0; iRow<nRows; iRow++) {
      inFrames[iRow] = minrtsFrameDataReadFitsRow(inFileName, 1+iRow, &ierr);
      if (ierr != 0) {
         printf("ERROR: %i opening input data table in %s\n", ierr, inFileName);
         printf("%s \n",strerror(errno));
         return 1;
      }
   }
   nx = inFrames[0]->nx[0];
   ny = inFrames[0]->ny[0];
   npix = nx * (long)ny;
/*
      get cross coupling coefficients
*/
   if (crossSpecified) {
      cross = minrtsFrameDataReadFitsRow(crossFileName,1,&ierr);
      if (ierr != 0) {
         printf("ERROR: %i getting kappa coeff from file %s\n", ierr, crossFileName);
         printf("%s \n", strerror(errno));
         return 1;
      }
      printf("   using kappa coefficients from  %s\n", crossFileName);
   }
/*
         loop through input frames;  clone output frames
         create a copy in scratch; interpolate in place with distortion
         correction into window 2 or 3 ; then copy into output frames
         Then multiply by cross coupling coeff
*/
   for (iRow=0; iRow<nRows; iRow++) {
      for (iFile = 0; iFile <2; iFile++ )  {
         outFrames[iFile][iRow] = minrtsFrameDataClone(inFrames[iRow]);
         for (iRegion = 2;iRegion <4; iRegion ++)  {
            scratch = minrtsFrameDataClone(inFrames[iRow]);
            midiDiffRectXYWindow(scratch, CoordDef, iRegion, &ierr);
            if (iFile==0) {    /*  A data */
               pixCpy(outFrames[iFile][iRow]->data[iRegion-1],scratch->data[3],npix);
               if (crossSpecified) 
                  for (iy=0;iy<ny;iy++) for (ix=0;ix<nx;ix++) 
                     outFrames[iFile][iRow]->data[iRegion-1][ix+iy*nx] *=
                        cross->data[iRegion-2][ix];
            } else {   /*  B data */
               pixCpy(outFrames[iFile][iRow]->data[iRegion-1], scratch->data[0], npix);
               if (crossSpecified) 
                  for (iy=0;iy<ny;iy++) for (ix=0;ix<nx;ix++) 
                     outFrames[iFile][iRow]->data[iRegion-1][ix+iy*nx] *=
                        cross->data[iRegion][ix];
            }
            minrtsFrameDataKill(scratch);
         } /* iRegion loop */
      } /* iFile loop */
   } /* iRow loop */
#ifdef RECAL
   for (i=0;i<4;i++) for(ix=0;ix<nx;ix++) for (iy=0;iy<ny;iy++) 
         Phot->data[i][ix+iy*nx] = Phot->data[i][ix+iy*nx]*cross->data[i][ix];
/*
      modify effective kappa coeff to make sum of Phot[0+2] = Target[1]
      modify effective kappa coeff to make sum of Phot[1+3] = Target[3]
*/
   for (ix=0;ix<nx;ix++) {
      p02 = 0.;
      t1  = 0.;
      p13 = 0.;
      t2  = 0.;
      for (iy=yBottom;iy<yTop;iy++) {
         i = ix + iy * nx;
         p02 = p02+Phot->data[0][i] + Phot->data[2][i];
         p13 = p13+Phot->data[1][i] + Phot->data[3][i];
         t1  = t1 +Target->data[1][i];
         t2  = t2 +Target->data[2][i];
      }
      if (p02 > 0.) t1 = t1/p02; else t1 = 1.;
      if (p13 > 0.) t2 = t2/p13;else t2 = 1.;
      for (iy=0;iy<ny;iy++) {
         i = ix + iy * nx;
         Phot->data[0][i] = Phot->data[0][i] * t1;
         Phot->data[2][i] = Phot->data[2][i] * t1;
         Phot->data[1][i] = Phot->data[1][i] * t2;
         Phot->data[3][i] = Phot->data[3][i] * t2;
      }
   }
#endif  /* RECAL */
/*
      make output files;  note that we would like to use minrtsFrameDataWrite
      to write output, but this routine does not allow copying of header information
      so we do it the hard way.  This involves reopening the input file to
      get the header information.
*/
   for (iFile = 0; iFile<2; iFile++) {
      strcpy(outFileName, outFileBase);
      if (iFile ==0) strcat (outFileName, AFileTail);
         else strcat (outFileName, BFileTail);
      fits_create_file (&outFile, outFileName, &ierr);
      if (ierr !=0) {
         fprintf (stderr, "ERROR opening fits output %s %i\n",outFileName, ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
      i = fits_open_file(&inFile, inFileName, READONLY, &ierr) ;
      i = fits_copy_header(inFile, outFile, &ierr);
      DetectorDef = read_imaging_detector_table(inFile, &ierr);
      odata       = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
      outData     = make_imaging_fdata_row(odata);
      for (iRow=0;iRow<nRows;iRow++) {
         for (iChan=0;iChan<4;iChan++) pixCpy (outData->data[iChan], 
            outFrames[iFile][iRow]->data[iChan],npix);
         i = write_imaging_fdata_table(outFile, odata, outData, iRow+1, &ierr);
      } /* iRow loop */
      i = close_imaging_fdata_table(outFile, odata, &ierr);
      kill_imaging_fdata_row(outData);
      printf("   successfully wrote output file %s\n", outFileName);
      i = fits_close_file(inFile, &ierr);
   } /* iFile loop */
      
   for (i=0;i<nRows;i++) minrtsFrameDataKill(inFrames[i]);
   for (iFile=0; iFile<2; iFile++) {
      for (i=0;i<nRows;i++) minrtsFrameDataKill(outFrames[iFile][i]);
      free(outFrames[iFile]);
   }
   free(inFrames);
   if (crossSpecified) minrtsFrameDataKill(cross);
   return 0;
}
